//---------------------------------------------------------------------------------------
// This file is part of the Lomse library.
// Copyright (c) 2010-present, Lomse Developers
//
// Licensed under the MIT license.
//
// See LICENSE and NOTICE.md files in the root directory of this source tree.
//---------------------------------------------------------------------------------------

#ifndef __LOMSE_BOX_SYSTEM_H__
#define __LOMSE_BOX_SYSTEM_H__

#include "lomse_basic.h"
#include "lomse_gm_basic.h"

//using namespace std;

namespace lomse
{

//forward declarations
class GmoBoxScorePage;
class TimeGridTable;
class GmMeasuresTable;
class GmoBoxSliceStaff;
class SystemLayouter;

//---------------------------------------------------------------------------------------
/** GmoBoxSystem represents a line of music in the printed score.
*/
class GmoBoxSystem : public GmoBox
{
protected:
	vector<GmoShapeStaff*> m_staffShapes;
	vector<int> m_firstStaff;       //index to first staff for each instrument
    TimeGridTable* m_pGridTable;
    int m_iPage;        //index of score page (0..n-1) in which this system is contained
    int m_iSystem;      //index of this system in the score (0..n-1)

	vector<int> m_iFirstMeasure;    //index to first measure, per instrument
    vector<int> m_nMeasures;        //number of measures in this system, per instrument
    LUnits m_dxFirstMeasure;        //shift from box left (virtual end barline of previous measure)

    //free vertical space at top and bottom
    LUnits m_uFreeAtTop = 0.0f;
    LUnits m_uFreeAtBottom = 0.0f;

public:
    GmoBoxSystem(ImoScore* pScore);
    ~GmoBoxSystem();

    //helpers for layout
    inline void set_top_limit(LUnits minLimit) { m_uFreeAtTop = minLimit - get_top(); }
    inline void set_bottom_limit(LUnits maxLimit) { m_uFreeAtBottom = get_bottom() - maxLimit; }
    inline LUnits get_free_space_at_top() { return m_uFreeAtTop; }
    inline LUnits get_free_space_at_bottom() { return m_uFreeAtBottom; }
    inline void set_free_space_at_top(LUnits space) { m_uFreeAtTop = space; }
    inline void set_free_space_at_bottom(LUnits space) { m_uFreeAtBottom = space; }

    /**  Move boxes and shapes to theirs final 'y' positions. */
    void reposition_slices_and_shapes(const std::vector<LUnits>& yOrgShifts,
                                      const std::vector<LUnits>& heights,
                                      const std::vector<LUnits>& barlinesHeight,
                                      const std::vector<std::vector<LUnits>>& relStaffTopPositions,
                                      LUnits bottomMarginIncr,
                                      SystemLayouter* pSysLayouter);

    //slices
	inline int get_num_slices() const { return (int)m_childBoxes.size(); }
    inline GmoBoxSlice* get_slice(int i) const { return (GmoBoxSlice*)m_childBoxes[i]; }
    GmoBoxSliceInstr* get_first_instr_slice(int iInstr);
    GmoBoxSliceInstr* find_instr_slice_at(LUnits x, LUnits y);
    GmoBoxSliceStaff* get_first_slice_staff_for(int iInstr, int iStaff);
    void reposition_slices(USize shift);
    void remove_free_space_at_bottom_and_adjust_slices();

    //grid table: xPositions/timepos
    inline void set_time_grid_table(TimeGridTable* pGridTable) { m_pGridTable = pGridTable; }
    inline TimeGridTable* get_time_grid_table() { return m_pGridTable; }
    TimeUnits start_time();
    TimeUnits end_time();
    LUnits get_x_for_note_rest_at_time(TimeUnits timepos);
    LUnits get_x_for_barline_at_time(TimeUnits timepos);

	//miscellaneous info
	inline int get_system_number() { return m_iSystem; }
    GmoShapeStaff* get_staff_shape(int absStaff);
    GmoShapeStaff* get_staff_shape(int iInstr, int iStaff);
    int instr_number_for_staff(int absStaff);
    int staff_number_for(int absStaff, int iInstr);
    inline void set_page_number(int iPage) { m_iPage = iPage; }
	inline int get_page_number() { return m_iPage; }

	//measures info
	/** Returns the index (0..n-1) for the measure that starts in this system, or
	    -1 if no measures in this instrument
    */
	int get_first_measure(int iInstr);

	/** Lomse considers two positions for barlines at the end of a system:
	    - the real position at end of previous system, and
	    - the virtual position at start of next system (the first
          available position after prolog).

        This method returns the virtual start position for the mesaure starting
        the system.
	*/
    inline LUnits get_start_measure_xpos() { return m_origin.x + m_dxFirstMeasure; }

    //Staff shapes
    GmoShapeStaff* add_staff_shape(GmoShapeStaff* pShape);
    void add_num_staves_for_instrument(int staves);
    inline vector<GmoShapeStaff*>& get_staff_shapes() { return m_staffShapes; }

    //hit tests related
    int staff_at(LUnits y);

    //helper. User API related
    int get_num_instruments();
    LUnits tenths_to_logical(Tenths value, int iInstr=0, int iStaff=0);

    //debug
    string dump_timegrid_table();
    string dump_measures_info();

protected:
    //building measures table
    friend class TimeSlice;
    void add_barline_info(int iMeasure, int iInstr);

    friend class SystemLayouter;
    inline void add_shift_to_start_measure(LUnits width) { m_dxFirstMeasure += width; }

    friend class GmoBoxScorePage;
    inline void set_system_number(int iSystem) { m_iSystem = iSystem; }

    //overrides
    void draw_box_bounds(Drawer* pDrawer, double xorg, double yorg, Color& color) override;

};



}   //namespace lomse

#endif      //__LOMSE_BOX_SYSTEM_H__
